/*
 * Driver for Ricoh RV5C386A RTC chip
 *
 * Copyright (c) 2007 Karl Brose
 * Strongly based on other RTC drivers by Broadcom.
 *
 * This program is free software. It is released under the terms of the
 * GNU General Public License 2.0.
 *
 * Copyright 2001-2004, Broadcom Corporation
 * All Rights Reserved.
 *
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * NOTE: 
 * This driver leverages lower level Linux I2C routines, for the 2-wire 
 * serial interface for RV5C386A.
 *
 * $Id: rtcrv5c386a.c,v 1.2 2007/10/14 19:24:04 nasmaker Exp $
*/

#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/mc146818rtc.h>
#include <linux/init.h>
#include <linux/time.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <asm/time.h>

#include <serbus.h>
#include <bcmrtc.h>

#define DEVICE_NAME		"RV5C386A"
#define DEVICE_SLAVE_ADDRESS	0x32	/* chip's slave address, see datasheet */

#define	DEVICE_ERR(s)	printk s

#define BCMDBG 0

#if BCMDBG
#define	DEVICE_TRACE(s)	printk s
#else
#define	DEVICE_TRACE(s)
#endif
#define ASSERT(x)	((x) ? (void)0 : printk("%s assert failed - line = %d\n", DEVICE_NAME, __LINE__ ))


extern unsigned long	(*rtc_get_time)(void);
extern int		(*rtc_set_time)(unsigned long);
extern void		to_tm(unsigned long tim, struct rtc_time * tm);

static struct i2c_driver rv5c386a_driver;

static unsigned short ignore[]		= {	I2C_CLIENT_END };

static unsigned short normal_addr[]	= {	DEVICE_SLAVE_ADDRESS,
						I2C_CLIENT_END };

static struct i2c_client_address_data
addr_data = {
	normal_i2c:		normal_addr,
	normal_i2c_range:	ignore,
	probe:			ignore,
	probe_range:		ignore,
	ignore:			ignore,
	ignore_range:		ignore,
	force:			ignore,
};

static struct i2c_client	rtc_client_mem;
static struct i2c_client*	rtc_client = NULL;


static int
rv5c386a_attach(struct i2c_adapter *adap, int addr, unsigned short flags, int kind)
{
	struct i2c_client *c = &rtc_client_mem;
	int ret = 0;

	strcpy(c->name, DEVICE_NAME);
	c->id		= rv5c386a_driver.id;
	c->flags	= 0;
	c->addr		= addr;
	c->adapter	= adap;
	c->driver	= &rv5c386a_driver;
	c->data		= NULL;

	if ((ret = i2c_smbus_write_quick(c, 0x0)) != 0) {
		DEVICE_ERR(("%s: Could not set control register, return value=%d\n", __FUNCTION__, ret));
		goto err0;
	}

	if ((ret  = i2c_attach_client(c)) != 0) {
		goto err0;
	}

	rtc_client = c;
#ifdef MODULE
	MOD_INC_USE_COUNT;
	i2c_inc_use_client(rtc_client);
#endif
	
err0:
	return ret;	
}


static int
rv5c386a_probe(struct i2c_adapter *adap)
{
	return i2c_probe(adap, &addr_data, rv5c386a_attach);
}


static int
rv5c386a_detach(struct i2c_client *client)
{
#ifdef MODULE
	i2c_dec_use_client(client);
	MOD_DEC_USE_COUNT;
#endif
	i2c_detach_client(client);
	rtc_client = NULL;
	return 0;
}


static int
rv5c386a_get_datetime(struct rtc_time *dt)
{
	struct i2c_client *client = rtc_client;
	unsigned char buf[7];

	if (client == NULL) {
		/* No Client, return error. */
		return -EIO;
	}

	buf[0] = i2c_smbus_read_byte_data(client, 0x00);
	buf[1] = i2c_smbus_read_byte_data(client, 0x10);
	buf[2] = i2c_smbus_read_byte_data(client, 0x20);
	buf[3] = i2c_smbus_read_byte_data(client, 0x30);
	buf[4] = i2c_smbus_read_byte_data(client, 0x40);
	buf[5] = i2c_smbus_read_byte_data(client, 0x50);
	buf[6] = i2c_smbus_read_byte_data(client, 0x60);

	DEVICE_TRACE(("%s: Year=%02x Month=%02x Day=%02x WeekDay=%02x Hour=%02x Minutes=%02x Seconds=%02x\n",
		__FUNCTION__,
		buf[6], buf[5], buf[4], buf[3], buf[2], buf[1], buf[0]));

	dt->tm_sec  = BCD_TO_BIN(buf[0]);
	dt->tm_min  = BCD_TO_BIN(buf[1]);
	dt->tm_hour = BCD_TO_BIN(buf[2]);
	dt->tm_wday = BCD_TO_BIN(buf[3]);
	dt->tm_mday = BCD_TO_BIN(buf[4]);
	dt->tm_mon  = BCD_TO_BIN(buf[5]);
	dt->tm_year = BCD_TO_BIN(buf[6]);

	return 0;
}

static int
rv5c386a_set_datetime(struct rtc_time *dt, int datetoo)
{
	struct i2c_client *client = rtc_client;
	unsigned char buf[8];	/* Include the register pointer in buf[0] */

	if (client == NULL) {
		/* No Client, return error. */
		return -EIO;
	}

	/* Point register to 0 */
	buf[0] = 0;
	buf[1] = BIN_TO_BCD(dt->tm_sec);
	buf[2] = BIN_TO_BCD(dt->tm_min);
	buf[3] = BIN_TO_BCD(dt->tm_hour);
	if (datetoo) {
        buf[4] = BIN_TO_BCD(dt->tm_wday);
        buf[5] = BIN_TO_BCD(dt->tm_mday);
        buf[6] = BIN_TO_BCD(dt->tm_mon);
        buf[7] = BIN_TO_BCD(dt->tm_year);
	}

	DEVICE_TRACE(("%s: Year=%02x Month=%02x Day=%02x WeekDay=%02x Hour=%02x Minutes=%02x Seconds=%02x\n",
		__FUNCTION__,
		buf[7], buf[6], buf[5], buf[4],  buf[3], buf[2], buf[1]));

	i2c_smbus_write_byte_data(client, 0x00, buf[1]);
	i2c_smbus_write_byte_data(client, 0x10, buf[2]);
	i2c_smbus_write_byte_data(client, 0x20, buf[3]);
	i2c_smbus_write_byte_data(client, 0x30, buf[4]);
	i2c_smbus_write_byte_data(client, 0x40, buf[5]);
	i2c_smbus_write_byte_data(client, 0x50, buf[6]);
	i2c_smbus_write_byte_data(client, 0x60, buf[7]);

	return 0;
}

static int
rv5c386a_command(struct i2c_client *client, unsigned int cmd, void *arg)
{
	int ret = 0;
	int lockFlag;

	if (client == NULL) {
		/* No Client, return error. */
		return -EIO;
	}

	if((lockFlag = bcm_serbus_lock(NULL)) < 0)
		return -ERESTARTSYS;

	switch (cmd) {
	case RTC_GETDATETIME:
		ret = rv5c386a_get_datetime(arg);
		break;

	case RTC_SETTIME:
		ret = rv5c386a_set_datetime(arg, 0);
		break;

	case RTC_SETDATETIME:
		ret = rv5c386a_set_datetime(arg, 1);
		break;

	default:
		ret = -EINVAL;
	}

	bcm_serbus_unlock(NULL, lockFlag);

	return ret;
}



int
rtcRV5C386A_setup(void)
{
	/*
	If the system is turned ON without the driver the RV5C386A gets
	in an error state, hence do this to clear the status once.
	*/
	struct i2c_client *client = rtc_client;
	int ret;

	if (client == NULL) {
		/* No Client, return error. */
		goto err0;
	}

	/* Initialize chip control registers. */
	/* Control register 1 */
	if ((ret = i2c_smbus_write_byte_data(client, 0xE0, 0x20) != 0)) {
		DEVICE_ERR(("%s: Could not set control register 1, return value=%d\n", __FUNCTION__, ret));
		goto err0;
	}
#if 0
	/* Control register 2 */
	if ((ret = i2c_smbus_write_byte_data(client, 0xF0, 0x0)) != 0) {
		DEVICE_ERR(("%s: Could not set status register 2, return value=%d\n", __FUNCTION__, ret));
		goto err0;
	}
#endif

	DEVICE_TRACE(("rtcRV5C386A_setup Done\n"));
	return 0;

err0:
	DEVICE_ERR(("rtcRV5C386A_setup Failed\n"));
	return -ENODEV;
}


static struct i2c_driver
rv5c386a_driver = {
	name:		DEVICE_NAME,
	id:		I2C_DRIVERID_EXP0,
	flags:		I2C_DF_NOTIFY,
	attach_adapter:	rv5c386a_probe,
	detach_client:	rv5c386a_detach,
	command:	rv5c386a_command
};

static bcmrtc_drvrinfo_t
rv5c386a_info = {
	name:			DEVICE_NAME,
	bcmrtc_get_datetime:	rv5c386a_get_datetime,
	bcmrtc_set_datetime:	rv5c386a_set_datetime
};


static __init int
rv5c386a_init(void)
{
	int ret = 0;

	if ((ret = i2c_add_driver(&rv5c386a_driver)) == 0) {
#ifdef MODULE
		if ((ret = rtcRV5C386A_setup()) == 0) {
			ret = bcmrtc_init(&rv5c386a_info);
		}
#endif
	}

	return ret;
}


static void __exit
rv5c386a_exit(void)
{
	bcmrtc_exit();
	i2c_del_driver(&rv5c386a_driver);
}

#ifdef MODULE
module_init(rv5c386a_init);
module_exit(rv5c386a_exit);
#endif
